/* cpp/fficpp.c
 * 2000-06-20 / lth
 *
 * Written by Lars Thomas Hansen (lth@cs.uoregon.edu)
 * Copyright 1996 The University of Oregon.
 *
 * These sources are freely redistributable with no restrictions whatsoever,
 * except that you may not remove this copyright notice.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "cpp.h"

static char *ffi_macrofile = 0;

static void dumpnlist( FILE *out );
static void printtok( Token *t, FILE *out );

void
setup_ffigen( char *filename )
{
  ffi_macrofile = filename;
}

void 
dump_ffigen_macros( void )
{
  FILE *out;
  
  if (!ffi_macrofile) 
    return;
  
  if ((out = fopen( ffi_macrofile, "w" )) == 0) {
    fprintf( stderr, "Can't create %s\n", ffi_macrofile );
    exit( 1 );
  }
  dumpnlist( out );
  fclose( out );
}

static void
dumpnlist( FILE *out )
{
  Nlist *np;
  Token *tp, *lim;
  void printtok( Token *, FILE * );

  for ( np=nlist_get(0) ; np ; np=nlist_get(1) )
    while (np) {
      if (np->flag == 1) {
	fprintf( out, "(macro ");

	/* In preliminary testing, a check for null didn't seem to be
           necessary, but it's nice to avoid segfaults when you can. C
           - what a pain. */
	if (np->source != NULL)
	  fprintf(out, "(\"%s\" %d)", np->source->filename, np->source->line);
	else
	  fprintf(out, "\"\" 0");

	fprintf(out, " \"%s", np->name);

	if (np->ap) {
	  fprintf( out, " ( " );
	  tp = np->ap->tp;
	  lim = np->ap->lp;
	  while (tp < lim) {
	    fprintf( out, "%s ", tp->t );
	    if (++tp != lim) 
	      fprintf( out, ", " );
	  }
	  fprintf( out, ")" );
	}
	fprintf( out, "\" \"" );
	if (np->vp) {
	  tp = np->vp->tp;
	  lim = np->vp->lp;
	  while (tp < lim) {
	    printtok( tp, out );
	    ++tp;
	    if (tp < lim)
	      putc( ' ', out );
	  }
	}
	fprintf( out, "\")\n" );
      }
      np = np->next;
    }
}

static void
printtok( Token *t, FILE *out )
{
  char *tbl[] = { 0, 0, 0, 0, 0, 0, "\n", " ", "##",
		  "==", "!=", "<=", ">=", "<<", ">>", "&&", "||", "++", "--",
		  "->", "[", "]", "(", ")", ".", "&", "*", "+", "-",
		  "~", "!", "/", "%", "<", ">", "^", "|", "?",
		  ":", "=", ",", "#", ";", "{", "}",
		  "+=", "-=", "*=", "/=", "%=", "^=", "<<=",
		  ">>=", "|=", "&=", "...",
		  "##", 0, 0, "-" };
  int i;

  if (tbl[t->type] == 0) {
    if (t->len > 0) {
      /* If the macro expands to something with quotes embedded in it,
         then the quotes need to be escaped */
      for (i = 0; i < t->len; i++) {
	if (t->t[i] == '"')
	  putc('\\', out);
	putc(t->t[i], out);
      }
    } else 
      fprintf( out, "XXX" );
  }
  else
    fprintf( out, "%s", tbl[t->type] );
}

/* eof */
