#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "cpp.h"


Includelist	includelist[NINCLUDE];

extern char	*objname;
int find_directory(char *filename);


/* Process an include directive.  If NEXT is set to 1 then, process as
   an include_next directive.  This causes the routine to look for the
   specified header file in the include directory which comes AFTER
   the directory of the header file currently being processed.
*/

void
doinclude(Tokenrow *trp, int next)
{
	char fname[256], iname[256];
	Includelist *ip;
	int angled, len, fd, i;
	int dir_found;

	trp->tp += 1;
	if (trp->tp>=trp->lp)
		goto syntax;
	if (trp->tp->type!=STRING && trp->tp->type!=LT) {
		len = trp->tp - trp->bp;
		expandrow(trp, "<include>");
		trp->tp = trp->bp+len;
	}
	if (trp->tp->type==STRING) {
		len = trp->tp->len-2;
		if (len > sizeof(fname) - 1)
			len = sizeof(fname) - 1;
		strncpy(fname, (char*)trp->tp->t+1, len);
		angled = 0;
	} else if (trp->tp->type==LT) {
		len = 0;
		trp->tp++;
		while (trp->tp->type!=GT) {
			if (trp->tp>trp->lp || len+trp->tp->len+2 >= sizeof(fname))
				goto syntax;
			strncpy(fname+len, (char*)trp->tp->t, trp->tp->len);
			len += trp->tp->len;
			trp->tp++;
		}
		angled = 1;
	} else
		goto syntax;
	trp->tp += 2;
	if (trp->tp < trp->lp || len==0)
		goto syntax;
	fname[len] = '\0';
	if (next) {
#ifdef FFI_DEBUG
	  fprintf(stderr,"include_next processing.\n");
	  fprintf(stderr,"fname = %s\n", fname);
	  fprintf(stderr,"cursource = %s\n", cursource->filename);
	  for (i=NINCLUDE-1; i>=0 && includelist[i].file != NULL; i--) {
	    fprintf(stderr,"includelist[%d] = %s\n", i, includelist[i].file);
	  }
#endif
	  /* Find the include directory of the current header file. */
	  dir_found = find_directory(cursource->filename);
	  if (dir_found < 0) {
	    error(FATAL, "Couldn't find include directory for current source file `%s`.", cursource->filename);
	  }
#ifdef FFI_DEBUG
	  fprintf(stderr,"dir_found = %d\n", dir_found);
#endif
	} else {
	  /* Regular include, start searching from the first include
             directory. */
	  dir_found = NINCLUDE;
	}
	if (fname[0]=='/') {
		fd = open(fname, 0);
		strcpy(iname, fname);
	} else for (fd = -1,i=dir_found-1; i>=0; i--) {
		ip = &includelist[i];
		if (ip->file==NULL || ip->deleted || (angled && ip->always==0))
			continue;
		if (strlen(fname)+strlen(ip->file)+2 > sizeof(iname))
			continue;
		strcpy(iname, ip->file);
		strcat(iname, "/");
		strcat(iname, fname);
		if ((fd = open(iname, 0)) >= 0)
			break;
	}
	if ( Mflag>1 || !angled&&Mflag==1 ) {
		write(1,objname,strlen(objname));
		write(1,iname,strlen(iname));
		write(1,"\n",1);
	}
#ifdef FFI_DEBUG
	fprintf(stderr,"iname = %s\n\n", iname);
#endif
	/* Why is this only 10 ? */
	if (fd >= 0) {
		if (++incdepth > 10)
			error(FATAL, "#include too deeply nested");
		setsource((char*)newstring((uchar*)iname, strlen(iname), 0), fd, NULL);
		genline();
	} else {
		trp->tp = trp->bp+2;
		error(ERROR, "Could not find include file %r", trp);
	}
	return;
syntax:
	error(ERROR, "Syntax error in #include");
	return;
}

/*
 * Generate a line directive for cursource
 */
void
genline(void)
{
	static Token ta = { UNCLASS };
	static Tokenrow tr = { &ta, &ta, &ta+1, 1 };
	uchar *p;

	ta.t = p = (uchar*)outp;
	strcpy((char*)p, "#line ");
	p += sizeof("#line ")-1;
	p = (uchar*)outnum((char*)p, cursource->line);
	*p++ = ' '; *p++ = '"';
	if (cursource->filename[0]!='/' && wd[0]) {
		strcpy((char*)p, wd);
		p += strlen(wd);
		*p++ = '/';
	}
	strcpy((char*)p, cursource->filename);
	p += strlen((char*)p);
	*p++ = '"'; *p++ = '\n';
	ta.len = (char*)p-outp;
	outp = (char*)p;
	tr.tp = tr.bp;
	puttokens(&tr);
}

void
setobjname(char *f)
{
	int n = strlen(f);
	objname = (char*)domalloc(n+5);
	strcpy(objname,f);
	if(objname[n-2]=='.'){
		strcpy(objname+n-1,"$O: ");
	}else{
		strcpy(objname+n,"$O: ");
	}
}

/* Take filename and try to see if it is contained in one of the
include directory paths. 

In general, this routine should ALWAYS succeed, since it is looking
for the path of a file which has already been included.  In other
words, the directory is in one of the include paths, or the program
would not have been able to load it in the first place.  */

int
find_directory(char *filename)
{
  char *dirname;
  int i;
  int index;

  index = -1;
  for (i=NINCLUDE-1; i>=0 && includelist[i].file != NULL; i--) {
    dirname = includelist[i].file;
    /* what to I do with the "." directory ?? */
    if (strncmp(dirname, filename, strlen(dirname)) == 0) {
      index = i;
    }
  }
  return index;
}

