#include "c.h"

static char rcsid[] = "$Id: ffigen.c,v 1.5 2001/01/21 01:02:37 briand Exp $";


#define OPEN    0x1
#define CLOSE   0x4

static indent_count = 1;

static void
indent(void)
{
  int i;

  printf("\n");
  for (i = 0; i < indent_count; i++) {
    printf(" ");
  }
  indent_count++;
}

static void
unindent(int n)
{
  indent_count -= n;
}

/*
A simple routine to print file info.  The filename and a dotted-pair
with the line and column info.  

The file information consists only of filename and line # for now.
Although the column nnumber is available, it's not used in the macro
processor and they should be consistent.  
*/

static void
print_file_info(Symbol s) {
  printf("(\"%s\" %d) ", (s->src.file ? s->src.file : ""), s->src.y);
}

void
examine_symbol(Symbol p)
{
  printf("p->name = \"%s\"\n", p->name);
  printf("p->scope = %d\n", p->scope);
  printf("p->sclass = %d\n", p->sclass);
  printf("p->up = \"%p\"\n", p->up);
  printf("p->type = \"%p\"\n", p->type);
  printf("p->u = \"%p\"\n", p->u);
  if (p->up != NULL) 
    examine_symbol(p->up);
}

void
examine_type(Type p)
{
  printf("\np->op = %d\n", p->op);
  printf("p->align = %d\n", p->align);
  printf("p->size = \"%p\"\n", p->size);
  printf("p->type = \"%p\"\n", p->type);
  printf("p->u = \"%p\"\n", p->u);
}


/* 
Write out hte bas type, i.e. the "attribute" setting.  

name is the name of the type being moidified, i.e. if the declaration is 

const int ...

Then name will be 'int'.

*/

void 
basetype(char *name, int attrs)
{
  printf("(");
  printf("%s", name);
  printf(" (");
  if (attrs == CONST)
    printf("const" );
  else if (attrs == VOLATILE)
    printf("volatile" );
  else if (attrs)
    printf("const volatile" );
  printf("))");
}

/*

print out the representation of a type.

When tdef = 1, we are processing a type and so should not print the
typdef type but print the full dereferenced type. */

void 
print_type(Type t, int attrs)
{
  Type *a;

  /* Handle special cases 

   Hey ! How the heck does this work. t is a structure, how can you
   test for equality ?? */

  if ((t->typedefsym != NULL)) {
    printf("(typedef \"%s\")", t->typedefsym->name);
    return;
  }
  
  if (t == chartype) basetype("char", attrs);
  else if (t == doubletype) basetype("double", attrs);
  else if (t == floattype) basetype("float", attrs);
  else if (t == inttype) basetype("int", attrs);
  else if (t == longdouble) basetype("long-double", attrs);
  else if (t == longtype) basetype("long", attrs);
  else if (t == longlong) basetype("long-long", attrs);
  else if (t == shorttype) basetype("short", attrs);
  else if (t == signedchar) basetype("signed-char", attrs);
  else if (t == unsignedchar) basetype("unsigned-char", attrs);
  else if (t == unsignedlong) basetype("unsigned-long", attrs);
  else if (t == unsignedlonglong) basetype("unsigned-long-long", attrs);
  else if (t == unsignedshort) basetype("unsigned-short", attrs);
  else if (t == unsignedtype) basetype("unsigned", attrs);
  else if (t == voidtype) basetype("void", attrs);

  switch (t->op) {
    case CHAR : 
    case SHORT : 
    case INT : 
    case UNSIGNED :
    case LONG : 
    case FLOAT : 
    case DOUBLE : 
    case VOID :
      return;
  }
  
  switch (t->op) {
    case ARRAY : 
      printf("(array %d ", t->size/t->type->size); 
      print_type(t->type, attrs); 
      printf(")");
      break;
    case ENUM :
      printf("(enum-ref \"%s\")", t->u.sym->name); 
      break;
    case STRUCT : 
      printf("(struct-ref \"%s\")", t->u.sym->name);
      break;
    case UNION : 
      printf("(union-ref \"%s\")", t->u.sym->name);
      break;
    case POINTER : 
      printf("(pointer "); 
      print_type(t->type, 0);
      printf(")");
      break;
    case FUNCTION : 
      printf("(function ");
      indent();
      /* prototype argument types */
      printf("(");
      a = t->u.f.proto;
      if (a) {
	while (*a) {
	  print_type(*a, 0);
	  if (*++a) 
	    printf(" ");
	}
      } else {
	printf("void ()");
      }
      printf(")");
      /* return type */
      unindent(1);
      indent();
      print_type(t->type, 0);
      unindent(1);
      printf(")");
      break;
    case CONST : 
    case VOLATILE :
    case CONST+VOLATILE :
      print_type(t->type, t->op);
      break;
    default:
      assert(0);
  }
}

void 
do_global(Symbol s, void *cl)
{
  char *class;
  class = "var";

  if (s->sclass == TYPEDEF) {
    class = "type";
  } else if ((s->sclass == STATIC || s->sclass == EXTERN || 
	      s->sclass == AUTO) &&
	     (s->type->op == FUNCTION))
    class = "function";
  /*  if (s->type && s->type->op == FUNCTION) printf("%d\n", s->sclass); */
  else if (s->sclass == ENUM)
    class = "enum-ident";

  printf("(");
  printf("%s ", class);
  print_file_info(s);
  indent();
  printf("\"%s\" ", s->name);
  unindent(1);
  if (s->sclass == ENUM)
    printf("%d", s->u.value);
  else {
    /* details, details. C pointer hell. */
    if (s->type != NULL) {
      indent();
      print_type(s->type, 0);
      unindent(1);
    }
  }
  if (strcmp(class, "type") != 0)
    switch (s->sclass) {
      case STATIC : printf(" (static)"); break;
      case EXTERN : printf(" (extern)"); break;
      default : 
	break;
    }
  printf(")");
  printf("\n");
}

/* Print the fields of union and struct types. 

Offset and size of the fields (including bit fields) is written out also.

So the format looks like :

<field>     -> (<name> <type> <offset> <size>)
<bit-field> -> (<name> <type> <offset> <size> 
                (<field-left> <field-right> <fieldsize> <fieldmask>))
*/

void 
print_fields( Symbol sym)
{
  Field f;
  int count = 0;

  f = sym->u.s.flist;
  indent();
  printf("(");
  while (f) {
    if (count > 0)
      indent();
    printf("(");
    printf("\"%s\" ", f->name);
    printf("(");
    if (f->lsb) {
      /* print a bit field and all of the offset information. */
      printf("bit-field ");
      print_type(f->type, 0);
      print(" %d %d (%d %d %d %d)", 
	    f->offset, f->type->size, (int)fieldleft(f), (int)fieldright(f), 
	    (int)fieldsize(f), (int)fieldmask(f));
    } else if (f->type != NULL) {
      /* print out a "normal" field plus the offset information. */
      printf("field ");
      print_type(f->type, 0);
      printf(" %d %d", f->offset, f->type->size);
    }
    printf("))");
    if (count > 0)
      unindent(1);
    count++;
    f = f->link;
  }
  printf(")");
  unindent(1);
}

void 
do_type(Symbol s, void *args)
{
  Symbol *p;

  if (!s->type) 
    return;

  if (! (s->type->op == UNION || s->type->op == STRUCT || s->type->op == ENUM))
    return;

  printf("(");

  if (s->type->op == UNION) {
    printf("union ");
    print_file_info(s);
    indent();
    printf("\"%s\" ", s->name);
    unindent(1);
  } else if (s->type->op == STRUCT) {
    printf("struct ");
    print_file_info(s);
    indent();
    printf("\"%s\" ", s->name);
    unindent(1);
  } else  if (s->type->op == ENUM) {
    printf("enum ");
    print_file_info(s);
    indent();
    printf("\"%s\" ", s->name);
    unindent(1);
    indent();
    printf("(");
    for (p = s->u.idlist ; *p ; p++) {
      printf("(");
      printf("\"%s\" %d", (*p)->name, (*p)->u.value);
      printf(")");
    }
    printf("))\n");
    unindent(1);
    return;
  }
  else
    return;

  print_fields(s);

  printf(")");
  printf("\n");
}


static void 
ffigen_segment(int n) 
{
}

static void 
ffigen_address(Symbol q, Symbol p, long n) 
{
}

static void
ffigen_blockbeg(Env *e) 
{
}

static void
ffigen_blockend(Env *e) 
{
}

static void
ffigen_defaddress(Symbol p) 
{
}

static void
ffigen_defconst(int suffix, int size, Value v) 
{
}

static void
ffigen_defstring(int len, char *str) 
{
}

static void
ffigen_defsymbol(Symbol p) 
{
#if 0
  printf("\nffigen_defsymbol.\n");
  examine_symbol(p);
#endif
}

static void
ffigen_emit(Node p) 
{
}

static void
ffigen_local(Symbol p) 
{
}

static void
ffigen_export(Symbol p) 
{
#if 0
  printf("ffigen_export.\n");
  printf("p->name = %s\n", p->name);
#endif
}

static void
ffigen_function(Symbol f, Symbol caller[], Symbol callee[], int ncalls) 
{
#if 0
  printf("ffigen_function.\n");
  printf("f->name = %s\n", f->name);
#endif
}

static Node
ffigen_gen(Node p) 
{
  return NULL;
}

static void
ffigen_global(Symbol p) 
{
#if 0
  printf("ffigen_global\n");
  do_global(p, NULL);
#endif
}

static void
ffigen_import(Symbol p) 
{
}

static void
ffigen_progbeg(int argc, char *argv[]) 
{
}

/* Invoked automagically by the compiler once the program has been
   digested. */

static void
ffigen_progend(void) 
{
  /* types and globals are global variables and so are accesible
     here. */
  foreach(types, GLOBAL, do_type, NULL);
  foreach(globals, GLOBAL, do_global, NULL);
}

static void
ffigen_space(int n) 
{
}

void 
ffigen_init(int argc, char *argv[]) 
{
  int i;
  static int inited = 0;

  if (inited)
    return;
  inited = 1;
#if 0
  printf("ffigen is inited.\n");
#endif
  for (i = 1; i < argc; i++)
    if (strcmp(argv[i], "-ffigen") == 0) {
      IR->address = ffigen_address;
      IR->blockbeg = ffigen_blockbeg;
      IR->blockend = ffigen_blockend;
      IR->defaddress = ffigen_defaddress;
      IR->defconst = ffigen_defconst;
      IR->defstring = ffigen_defstring;
      IR->defsymbol = ffigen_defsymbol;
      IR->emit = ffigen_emit;
      IR->export = ffigen_export;
      IR->function = ffigen_function;
      IR->gen = ffigen_gen;
      IR->global = ffigen_global;
      IR->import = ffigen_import;
      IR->local = ffigen_local;
      IR->progbeg = ffigen_progbeg;
      IR->progend = ffigen_progend;
      IR->segment = ffigen_segment;
      IR->space = ffigen_space;

      IR->stabblock = 0;
      IR->stabend = 0;
      IR->stabfend = 0;
      IR->stabinit = 0;
      IR->stabline = 0;
      IR->stabsym = 0;
      IR->stabtype = 0;

      IR->wants_dag = 0;
      prunetemps = 0;	/* pass2 prunes useless temps */
      assignargs = 0;	/* pass2 generates caller to callee assignments */
    }
}
